==> Saxon-HE-12.4-sources.jar: net/sf/saxon/event/CheckSumFilter.java <==
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2018-2023 Saxonica Limited
// This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
// If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
// This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

package net.sf.saxon.event;

import net.sf.saxon.s9api.Location;
import net.sf.saxon.om.*;
import net.sf.saxon.str.UnicodeString;
import net.sf.saxon.trans.XPathException;
import net.sf.saxon.transpile.CSharpReplaceBody;
import net.sf.saxon.type.SchemaType;
import net.sf.saxon.value.Whitespace;

/**
 * A filter to go on a Receiver pipeline and calculate a checksum of the data passing through the pipeline.
 * Optionally the filter will also check any checksum (represented by a processing instruction with name
 * SIGMA) found in the file.
 *
 * <p>The checksum takes account of element, attribute, and text nodes only. The order of attributes
 * within an element makes no difference.</p>
 */

public class CheckSumFilter extends ProxyReceiver {

    private final static boolean DEBUG = false;

    private int checksum = 0;
    private int sequence = 0;
    private boolean checkExistingChecksum = false;
    private boolean checksumCorrect = false;
    private boolean checksumFound = false;

    public final static String SIGMA = "\u03A3";

    public CheckSumFilter(Receiver nextReceiver) {
        super(nextReceiver);
    }

    /**
     * Ask the filter to check any existing checksums found in the file
     * @param check true if existing checksums are to be checked
     */

    public void setCheckExistingChecksum(boolean check) {
        this.checkExistingChecksum = check;
    }

    @CSharpReplaceBody(code="")
    private static void trace(String message) {
        if (DEBUG) {
            System.err.println(message);
        }
    }

    @Override
    public void startDocument(int properties) throws XPathException {
        trace("CHECKSUM - START DOC");
        super.startDocument(properties);
    }

    /**
     * Append an arbitrary item (node or atomic value) to the output
     * @param item           the item to be appended
     * @param locationId     the location of the calling instruction, for diagnostics
     * @param copyNamespaces if the item is an element node, this indicates whether its namespaces
*                       need to be copied. Values are {@link ReceiverOption#ALL_NAMESPACES};
     *                            the default (0) means
     */
    @Override
    public void append(Item item, Location locationId, int copyNamespaces) throws XPathException {
        checksum ^= hash(item.toString(), sequence++);
        trace("After append: " + Integer.toHexString(checksum));
        super.append(item, locationId, copyNamespaces);
    }

    /**
     * Character data
     */
    @Override
    public void characters(UnicodeString chars, Location locationId, int properties) throws XPathException {
        if (!Whitespace.isAllWhite(chars)) {
            checksum ^= hash(chars.toString(), sequence++);
            trace("After characters " + chars + ": " + Integer.toHexString(checksum));
        }
        super.characters(chars, locationId, properties);
    }

    /**
     * Notify the start of an element
     */
    @Override
    public void startElement(NodeName elemName, SchemaType type,
                             AttributeMap attributes, NamespaceMap namespaces,
                             Location location, int properties) throws XPathException {
        checksum ^= hash(elemName, sequence++);
        trace("After startElement " + elemName.getDisplayName() + ": " + checksum);
        checksumCorrect = false;
        for (AttributeInfo att : attributes) {
            checksum ^= hash(att.getNodeName(), sequence);
            checksum ^= hash(att.getValue(), sequence);
            if (DEBUG) {
                System.err.println("After attribute " + att.getNodeName().getDisplayName() + ": " + checksum +
                                           "(" + hash(att.getNodeName(), sequence) + "," + hash(att.getValue(), sequence) + ")");
            }
        }
        super.startElement(elemName, type, attributes, namespaces, location, properties);
    }

    /**
     * End of element
     */
    @Override
    public void endElement() throws XPathException {
        checksum ^= 1;
        trace("After endElement: " + checksum);
        super.endElement();
    }

    /**
     * Processing Instruction
     */
    @Override
    public void processingInstruction(String target, UnicodeString data, Location locationId, int properties) throws XPathException {
        if (target.equals(SIGMA)) {
            checksumFound = true;
            if (checkExistingChecksum) {
                try {
                    int found = (int) Long.parseLong("0" + data, 16);
                    checksumCorrect = found == checksum;
                } catch (NumberFormatException e) {
                    checksumCorrect = false;
                }
            }
        }
        super.processingInstruction(target, data, locationId, properties);
    }

    /**
     * Ask whether a checksum has been found
     * @return true if a checksum processing instruction has been found (whether or not the checksum was correct)
     */

    public boolean isChecksumFound() {
        return checksumFound;
    }

    /**
     * Get the accumulated checksum
     * @return the checksum of the events passed through the filter so far.
     */

    public int getChecksum() {
        return checksum;
    }

    /**
     * Ask if a correct checksum has been found in the file
     * @return true if a checksum has been found, if its value matches, and if no significant data has been encountered
     * after the checksum
     */

    public boolean isChecksumCorrect() {
        return checksumCorrect || "skip".equals(System.getProperty("saxon-checksum"));
    }

    private int hash(String s, int sequence) {
        int h = sequence<<8;
        for (int i=0; i<s.length(); i++) {
            h = (h<<1) + s.charAt(i);
        }
        return h;
    }

    private int hash(NodeName n, int sequence) {
        //System.err.println("hash(" + n.getLocalPart() + ") " + hash(n.getLocalPart(), sequence) + "/" + hash(n.getURI(), sequence));
        return hash(n.getLocalPart(), sequence) ^ hash(n.getNamespaceUri().toString(), sequence);
    }
}

==> Saxon-HE-12.5-sources.jar: net/sf/saxon/event/CheckSumFilter.java <==
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2018-2023 Saxonica Limited
// This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
// If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
// This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

package net.sf.saxon.event;

import net.sf.saxon.lib.DigestMaker;
import net.sf.saxon.om.*;
import net.sf.saxon.s9api.Location;
import net.sf.saxon.s9api.QName;
import net.sf.saxon.str.UnicodeString;
import net.sf.saxon.trans.XPathException;
import net.sf.saxon.transpile.CSharpReplaceBody;
import net.sf.saxon.type.SchemaType;
import net.sf.saxon.value.Whitespace;

import java.util.Arrays;
import java.util.HashMap;

/**
 * A filter to go on a Receiver pipeline and calculate a checksum of the data passing through the pipeline.
 * Optionally the filter will also check any checksum (represented by a processing instruction with name
 * SIGMA) found in the file.
 *
 * <p>The checksum takes account of element, attribute, and text nodes only. The order of attributes
 * within an element makes no difference.</p>
 */

public class CheckSumFilter extends ProxyReceiver {
    private final static boolean DEBUG = false;
    private DigestMaker digest = null;
    private int checksum = 0;
    private int sequence = 0;
    private boolean checkExistingChecksum = false;
    private boolean checksumCorrect = false;
    private boolean checksumFound = false;
    private boolean digestCorrect = false;
    private boolean digestFound = false;
    private boolean requireDigest = false;
    private boolean rootElement = true;

    public final static String SIGMA = "Σ";
    public final static String SIGMA2 = "Σ2";

    public CheckSumFilter(Receiver nextReceiver) {
        super(nextReceiver);
        rootElement = true;
        digest = new DigestMaker();
    }

    /**
     * Ask the filter to check any existing checksums found in the file
     * @param check true if existing checksums are to be checked
     */

    public void setCheckExistingChecksum(boolean check) {
        this.checkExistingChecksum = check;
    }

    @CSharpReplaceBody(code="")
    private static void trace(String message) {
        if (DEBUG) {
            System.err.println(message);
        }
    }

    @Override
    public void startDocument(int properties) throws XPathException {
        trace("CHECKSUM - START DOC");
        super.startDocument(properties);
    }

    @Override
    public void endDocument() throws XPathException {
        /*
        System.err.println("Σ ::= " + BMPString.of(Integer.toHexString(checksum)));
        System.err.println(digestbuffer.toString());
         */
        nextReceiver.endDocument();
    }

    /**
     * Append an arbitrary item (node or atomic value) to the output
     * @param item           the item to be appended
     * @param locationId     the location of the calling instruction, for diagnostics
     * @param copyNamespaces if the item is an element node, this indicates whether its namespaces
*                       need to be copied. Values are {@link ReceiverOption#ALL_NAMESPACES};
     *                            the default (0) means
     */
    @Override
    public void append(Item item, Location locationId, int copyNamespaces) throws XPathException {
        checksum ^= hash(item.toString(), sequence++);
        trace("After append: " + Integer.toHexString(checksum));
        super.append(item, locationId, copyNamespaces);
    }

    /**
     * Character data
     */
    @Override
    public void characters(UnicodeString chars, Location locationId, int properties) throws XPathException {
        if (!Whitespace.isAllWhite(chars)) {
            checksum ^= hash(chars.toString(), sequence++);
            trace("After characters " + chars + ": " + Integer.toHexString(checksum));
        }
        super.characters(chars, locationId, properties);
    }

    /**
     * Notify the start of an element
     */
    @Override
    public void startElement(NodeName elemName, SchemaType type,
                             AttributeMap attributes, NamespaceMap namespaces,
                             Location location, int properties) throws XPathException {
        checksum ^= hash(elemName, sequence++);
        trace("After startElement " + elemName.getDisplayName() + ": " + checksum);
        checksumCorrect = false;

        if (rootElement) {
            rootElement = false;

            boolean scm_schema = elemName.getNamespaceUri() == NamespaceUri.of("http://ns.saxonica.com/schema-component-model")
                    && "schema".equals(elemName.getLocalPart());

            // A digest is required for version 12.5+
            String version = attributes.getValue("saxonVersion");
            // No sneaky deleting the version to avoid the check. Except SCM files don't have a version...
            requireDigest = version == null && !scm_schema;
            if (version != null) {
                String minorVersion = "x"; // cause number format exception
                int dpos = version.indexOf('.');
                if (dpos > 0) {
                    minorVersion = version.substring(dpos+1);
                    version = version.substring(0, dpos);
                }
                try {
                    int majorVersion = Integer.parseInt(version);
                    if (majorVersion > 12) {
                        requireDigest = true;
                    } else if (majorVersion == 12) {
                        requireDigest = Integer.parseInt(minorVersion) >= 5;
                    }
                } catch (NumberFormatException e) {
                    requireDigest = true;
                }
            }
        }

        // Need these in lexicographic order for the cryptographic hash.
        // I also want to assure that the current checksum works. And I want
        // to use the hash() function as the common place for collecting data
        // for both. That means a bit of fiddling around here (extra fiddling
        // because NodeName doesn't implement Comparable).

        HashMap<String,NodeName> namemap = new HashMap<>();
        HashMap<String,String> attrmap = new HashMap<>();
        String[] names = new String[attributes.size()];
        int index = 0;
        for (AttributeInfo att : attributes) {
            String key = att.getNodeName().getLocalPart() + att.getNodeName().getNamespaceUri();
            attrmap.put(key, att.getValue());
            namemap.put(key, att.getNodeName());
            names[index++] = key;
        }

        Arrays.sort(names);
        for (String key : names) {
            NodeName name = namemap.get(key);
            String value = attrmap.get(key);
            checksum ^= hash(name, sequence);
            checksum ^= hash(value, sequence);
            if (DEBUG) {
                System.err.println("After attribute " + name.getDisplayName() + ": " + checksum +
                        "(" + hash(name, sequence) + "," + hash(value, sequence) + ")");
            }
        }

        super.startElement(elemName, type, attributes, namespaces, location, properties);
    }

    /**
     * End of element
     */
    @Override
    public void endElement() throws XPathException {
        checksum ^= 1;
        trace("After endElement: " + checksum);
        super.endElement();
    }

    /**
     * Processing Instruction
     */
    @Override
    public void processingInstruction(String target, UnicodeString data, Location locationId, int properties) throws XPathException {
        if (target.equals(SIGMA)) {
            checksumFound = true;
            if (checkExistingChecksum) {
                try {
                    int found = (int) Long.parseLong("0" + data, 16);
                    checksumCorrect = found == checksum;
                } catch (NumberFormatException e) {
                    checksumCorrect = false;
                }
                if (data.toString().equals(getDigest())) {
                    // This case represents some point in the future when we've
                    // abandoned the checksum and the digest is stored in SIGMA
                    digestFound = true;
                    digestCorrect = true;
                    checksumCorrect = true; // digest trumps checksum
                }
            }
        }
        if (target.equals(SIGMA2)) {
            digestFound = true;
            if (checkExistingChecksum) {
                digestCorrect = data.toString().equals(getDigest());
            }
        }
        super.processingInstruction(target, data, locationId, properties);
    }

    /**
     * Ask whether a checksum has been found
     * @return true if a checksum processing instruction has been found (whether or not the checksum was correct)
     */

    public boolean isChecksumFound() {
        return checksumFound;
    }

    public boolean isDigestFound() {
        return digestFound;
    }

    /**
     * Get the accumulated checksum
     * @return the checksum of the events passed through the filter so far.
     */

    public int getChecksum() {
        return checksum;
    }

    /**
     * Get the computed hash
     * @return the hash
     */
    public String getDigest() {
        return digest.getDigest();
    }

    /**
     * Ask if a correct checksum has been found in the file
     * @return true if a checksum has been found, if its value matches, and if no significant data has been encountered
     * after the checksum
     */

    public boolean isChecksumCorrect() {
        if (requireDigest && !digestCorrect) {
            return false;
        }
        return checksumCorrect || "skip".equals(System.getProperty("saxon-checksum"));
    }

    private int hash(String s, int sequence) {
        //System.err.printf("%d %d %s%n", sequence, s.length(), s);
        digest.update(sequence);
        digest.update(s);

        int h = sequence<<8;
        for (int i=0; i<s.length(); i++) {
            h = (h<<1) + s.charAt(i);
        }
        return h;
    }

    private int hash(NodeName n, int sequence) {
        //System.err.println("hash(" + n.getLocalPart() + ") " + hash(n.getLocalPart(), sequence) + "/" + hash(n.getURI(), sequence));
        return hash(n.getLocalPart(), sequence) ^ hash(n.getNamespaceUri().toString(), sequence);
    }
}
